from __future__ import annotations

import os

import ulid


def test_bench_new_crypto(benchmark) -> None:  # type: ignore[no-untyped-def]
    benchmark(ulid.new, ulid.now(), os.urandom)


def test_bench_new_monotonic_crypto(benchmark) -> None:  # type: ignore[no-untyped-def]
    entropy = ulid.monotonic(os.urandom, 0)
    benchmark(ulid.new, ulid.now(), entropy)


def test_bench_make(benchmark) -> None:  # type: ignore[no-untyped-def]
    benchmark(ulid.make)


def test_bench_parse(benchmark) -> None:  # type: ignore[no-untyped-def]
    s = ulid.make().string()
    benchmark(ulid.parse, s)


def test_bench_parse_strict(benchmark) -> None:  # type: ignore[no-untyped-def]
    s = ulid.make().string()
    benchmark(ulid.parse_strict, s)


def test_bench_parse_prefixed(benchmark) -> None:  # type: ignore[no-untyped-def]
    s = ulid.make().prefixed("user")
    benchmark(ulid.parse_prefixed, s)


def test_bench_string(benchmark) -> None:  # type: ignore[no-untyped-def]
    uid = ulid.make()
    benchmark(uid.string)


def test_bench_prefixed(benchmark) -> None:  # type: ignore[no-untyped-def]
    uid = ulid.make()
    benchmark(uid.prefixed, "user")


def test_bench_marshal_text(benchmark) -> None:  # type: ignore[no-untyped-def]
    uid = ulid.make()
    benchmark(uid.marshal_text)


def test_bench_marshal_binary(benchmark) -> None:  # type: ignore[no-untyped-def]
    uid = ulid.make()
    benchmark(uid.marshal_binary)


def test_bench_marshal_json(benchmark) -> None:  # type: ignore[no-untyped-def]
    uid = ulid.make()
    benchmark(uid.marshal_json)


def test_bench_compare(benchmark) -> None:  # type: ignore[no-untyped-def]
    a = ulid.make()
    b = ulid.make()
    benchmark(a.compare, b)


def test_bench_now(benchmark) -> None:  # type: ignore[no-untyped-def]
    benchmark(ulid.now)


def test_bench_time(benchmark) -> None:  # type: ignore[no-untyped-def]
    uid = ulid.make()
    benchmark(uid.time)


def test_bench_generator_make(benchmark) -> None:  # type: ignore[no-untyped-def]
    gen = ulid.new_generator()
    benchmark(gen.make)


def test_bench_generator_make_node_id(benchmark) -> None:  # type: ignore[no-untyped-def]
    gen = ulid.new_generator(ulid.with_node_id(42))
    benchmark(gen.make)


def test_bench_generator_make_prefixed(benchmark) -> None:  # type: ignore[no-untyped-def]
    gen = ulid.new_generator(ulid.with_node_id(1), ulid.with_prefix("txn"))
    benchmark(gen.make_prefixed)



from __future__ import annotations

import os
import threading
from collections.abc import Callable

from ulid.monotonic import MonotonicEntropy, monotonic
from ulid.ulid import ULID, new, now


class Generator:
    __slots__ = ("_mu", "_entropy", "_node_id", "_has_node", "_prefix")

    def __init__(self) -> None:
        self._mu = threading.Lock()
        self._entropy: MonotonicEntropy = monotonic(os.urandom, 0)
        self._node_id: int = 0
        self._has_node: bool = False
        self._prefix: str = ""

    def make(self) -> ULID:
        with self._mu:
            uid = new(now(), self._entropy)
            if self._has_node:
                buf = bytearray(uid._data)
                buf[6] = (self._node_id >> 8) & 0xFF
                buf[7] = self._node_id & 0xFF
                uid = ULID(bytes(buf))
        return uid

    def make_prefixed(self, prefix: str = "") -> str:
        uid = self.make()
        p = prefix if prefix else self._prefix
        if not p:
            raise RuntimeError("ulid: no prefix specified")
        return uid.prefixed(p)

    def new(self, ms: int) -> ULID:
        with self._mu:
            uid = new(ms, self._entropy)
            if self._has_node:
                buf = bytearray(uid._data)
                buf[6] = (self._node_id >> 8) & 0xFF
                buf[7] = self._node_id & 0xFF
                uid = ULID(bytes(buf))
        return uid

    def node_id(self) -> tuple[int, bool]:
        return self._node_id, self._has_node


GeneratorOption = Callable[[Generator], None]


def with_node_id(node_id: int) -> GeneratorOption:
    def apply(g: Generator) -> None:
        g._node_id = node_id
        g._has_node = True

    return apply


def with_entropy(reader: Callable[[int], bytes]) -> GeneratorOption:
    def apply(g: Generator) -> None:
        g._entropy = monotonic(reader, 0)

    return apply


def with_prefix(prefix: str) -> GeneratorOption:
    def apply(g: Generator) -> None:
        g._prefix = prefix

    return apply


def new_generator(*opts: GeneratorOption) -> Generator:
    g = Generator()
    for opt in opts:
        opt(g)
    return g



from __future__ import annotations

import builtins
import json
import os
import time as _time
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from collections.abc import Callable

    from ulid.monotonic import LockedMonotonicReader, MonotonicEntropy


class ULIDError(Exception):
    pass


class ErrDataSize(ULIDError):
    def __init__(self) -> None:
        super().__init__("ulid: bad data size when unmarshaling")


class ErrInvalidCharacters(ULIDError):
    def __init__(self) -> None:
        super().__init__("ulid: bad data characters when unmarshaling")


class ErrBufferSize(ULIDError):
    def __init__(self) -> None:
        super().__init__("ulid: bad buffer size when marshaling")


class ErrBigTime(ULIDError):
    def __init__(self) -> None:
        super().__init__("ulid: timestamp too big")


class ErrOverflow(ULIDError):
    def __init__(self) -> None:
        super().__init__("ulid: overflow when unmarshaling")


class ErrMonotonicOverflow(ULIDError):
    def __init__(self) -> None:
        super().__init__("ulid: monotonic entropy overflow")


class ErrScanValue(ULIDError):
    def __init__(self) -> None:
        super().__init__("ulid: source value must be a string or byte slice")


class ErrInvalidPrefix(ULIDError):
    def __init__(self) -> None:
        super().__init__("ulid: invalid prefix format")


ENCODED_SIZE = 26
BINARY_SIZE = 16
MAX_TIME = (1 << 48) - 1

_ENCODING = "0123456789abcdefghjkmnpqrstvwxyz"

# 256-byte decode lookup table. 0xFF = invalid character.
_DEC = bytes(
    [
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x00
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x08
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x10
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x18
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x20
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x28
        0x00,
        0x01,
        0x02,
        0x03,
        0x04,
        0x05,
        0x06,
        0x07,  # '0'-'7'
        0x08,
        0x09,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # '8'-'9'
        0xFF,
        0x0A,
        0x0B,
        0x0C,
        0x0D,
        0x0E,
        0x0F,
        0x10,  # 'A'-'G'
        0x11,
        0xFF,
        0x12,
        0x13,
        0xFF,
        0x14,
        0x15,
        0xFF,  # 'H'-'O'
        0x16,
        0x17,
        0x18,
        0x19,
        0x1A,
        0xFF,
        0x1B,
        0x1C,  # 'P'-'W'
        0x1D,
        0x1E,
        0x1F,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 'X'-'Z'
        0xFF,
        0x0A,
        0x0B,
        0x0C,
        0x0D,
        0x0E,
        0x0F,
        0x10,  # 'a'-'g'
        0x11,
        0xFF,
        0x12,
        0x13,
        0xFF,
        0x14,
        0x15,
        0xFF,  # 'h'-'o'
        0x16,
        0x17,
        0x18,
        0x19,
        0x1A,
        0xFF,
        0x1B,
        0x1C,  # 'p'-'w'
        0x1D,
        0x1E,
        0x1F,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 'x'-'z'
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x80
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x88
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x90
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0x98
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xA0
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xA8
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xB0
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xB8
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xC0
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xC8
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xD0
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xD8
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xE0
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xE8
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xF0
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,  # 0xF8
    ]
)

EntropyReader = Union["Callable[[int], bytes]", "MonotonicEntropy", "LockedMonotonicReader", None]


class ULID:
    __slots__ = ("_data",)

    def __init__(self, data: builtins.bytes = b"\x00" * 16) -> None:
        if len(data) != BINARY_SIZE:
            raise ErrDataSize()
        self._data = bytes(data)

    def __str__(self) -> str:
        return self.string()

    def __repr__(self) -> str:
        return f"ULID({self.string()!r})"

    def __bytes__(self) -> builtins.bytes:
        return self._data

    def __int__(self) -> int:
        return int.from_bytes(self._data, "big")

    def __hash__(self) -> int:
        return hash(self._data)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ULID):
            return self._data == other._data
        return NotImplemented

    def __ne__(self, other: object) -> bool:
        if isinstance(other, ULID):
            return self._data != other._data
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, ULID):
            return self._data < other._data
        return NotImplemented

    def __le__(self, other: object) -> bool:
        if isinstance(other, ULID):
            return self._data <= other._data
        return NotImplemented

    def __gt__(self, other: object) -> bool:
        if isinstance(other, ULID):
            return self._data > other._data
        return NotImplemented

    def __ge__(self, other: object) -> bool:
        if isinstance(other, ULID):
            return self._data >= other._data
        return NotImplemented

    def string(self) -> str:
        return _encode(self._data)

    def prefixed(self, prefix: str) -> str:
        return prefix + "_" + self.string()

    def bytes(self) -> builtins.bytes:
        return self._data

    def time(self) -> int:
        d = self._data
        return d[5] | (d[4] << 8) | (d[3] << 16) | (d[2] << 24) | (d[1] << 32) | (d[0] << 40)

    def timestamp(self) -> datetime:
        return time(self.time())

    def entropy(self) -> builtins.bytes:
        return self._data[6:]

    def is_zero(self) -> bool:
        return self._data == b"\x00" * 16

    def compare(self, other: ULID) -> int:
        if self._data < other._data:
            return -1
        if self._data > other._data:
            return 1
        return 0

    def set_time(self, ms: int) -> ULID:
        if ms > MAX_TIME:
            raise ErrBigTime()
        buf = bytearray(self._data)
        buf[0] = (ms >> 40) & 0xFF
        buf[1] = (ms >> 32) & 0xFF
        buf[2] = (ms >> 24) & 0xFF
        buf[3] = (ms >> 16) & 0xFF
        buf[4] = (ms >> 8) & 0xFF
        buf[5] = ms & 0xFF
        return ULID(bytes(buf))

    def set_entropy(self, e: builtins.bytes) -> ULID:
        if len(e) != 10:
            raise ErrDataSize()
        buf = bytearray(self._data)
        buf[6:] = e
        return ULID(bytes(buf))

    def marshal_binary(self) -> builtins.bytes:
        return self._data

    def marshal_text(self) -> builtins.bytes:
        return self.string().encode("ascii")

    def marshal_json(self) -> builtins.bytes:
        return json.dumps(self.string()).encode("ascii")

    @staticmethod
    def unmarshal_binary(data: builtins.bytes) -> ULID:
        if len(data) != BINARY_SIZE:
            raise ErrDataSize()
        return ULID(data)

    @staticmethod
    def unmarshal_text(data: builtins.bytes | str) -> ULID:
        if isinstance(data, str):
            return _parse(data, False)
        return _parse(data.decode("ascii"), False)

    @staticmethod
    def unmarshal_json(data: builtins.bytes | str) -> ULID:
        raw = data if isinstance(data, str) else data.decode("ascii")
        if len(raw) < 2 or raw[0] != '"' or raw[-1] != '"':
            raise ErrDataSize()
        return _parse(raw[1:-1], False)


def _encode(d: bytes) -> str:
    enc = _ENCODING
    # Timestamp (6 bytes -> 10 characters) - fully unrolled
    c0 = enc[(d[0] & 224) >> 5]
    c1 = enc[d[0] & 31]
    c2 = enc[(d[1] & 248) >> 3]
    c3 = enc[((d[1] & 7) << 2) | ((d[2] & 192) >> 6)]
    c4 = enc[(d[2] & 62) >> 1]
    c5 = enc[((d[2] & 1) << 4) | ((d[3] & 240) >> 4)]
    c6 = enc[((d[3] & 15) << 1) | ((d[4] & 128) >> 7)]
    c7 = enc[(d[4] & 124) >> 2]
    c8 = enc[((d[4] & 3) << 3) | ((d[5] & 224) >> 5)]
    c9 = enc[d[5] & 31]
    # Entropy (10 bytes -> 16 characters) - fully unrolled
    c10 = enc[(d[6] & 248) >> 3]
    c11 = enc[((d[6] & 7) << 2) | ((d[7] & 192) >> 6)]
    c12 = enc[(d[7] & 62) >> 1]
    c13 = enc[((d[7] & 1) << 4) | ((d[8] & 240) >> 4)]
    c14 = enc[((d[8] & 15) << 1) | ((d[9] & 128) >> 7)]
    c15 = enc[(d[9] & 124) >> 2]
    c16 = enc[((d[9] & 3) << 3) | ((d[10] & 224) >> 5)]
    c17 = enc[d[10] & 31]
    c18 = enc[(d[11] & 248) >> 3]
    c19 = enc[((d[11] & 7) << 2) | ((d[12] & 192) >> 6)]
    c20 = enc[(d[12] & 62) >> 1]
    c21 = enc[((d[12] & 1) << 4) | ((d[13] & 240) >> 4)]
    c22 = enc[((d[13] & 15) << 1) | ((d[14] & 128) >> 7)]
    c23 = enc[(d[14] & 124) >> 2]
    c24 = enc[((d[14] & 3) << 3) | ((d[15] & 224) >> 5)]
    c25 = enc[d[15] & 31]
    return (
        c0
        + c1
        + c2
        + c3
        + c4
        + c5
        + c6
        + c7
        + c8
        + c9
        + c10
        + c11
        + c12
        + c13
        + c14
        + c15
        + c16
        + c17
        + c18
        + c19
        + c20
        + c21
        + c22
        + c23
        + c24
        + c25
    )


def _parse(s: str, strict: bool) -> ULID:
    if len(s) != ENCODED_SIZE:
        raise ErrDataSize()

    v = s.encode("ascii") if isinstance(s, str) else s
    dec = _DEC

    if dec[v[0]] > 7:
        raise ErrOverflow()

    if strict:
        for i in range(ENCODED_SIZE):
            if dec[v[i]] == 0xFF:
                raise ErrInvalidCharacters()

    # Timestamp (10 characters -> 6 bytes) - fully unrolled
    b0 = (dec[v[0]] << 5) | dec[v[1]]
    b1 = (dec[v[2]] << 3) | (dec[v[3]] >> 2)
    b2 = (dec[v[3]] << 6) | (dec[v[4]] << 1) | (dec[v[5]] >> 4)
    b3 = (dec[v[5]] << 4) | (dec[v[6]] >> 1)
    b4 = (dec[v[6]] << 7) | (dec[v[7]] << 2) | (dec[v[8]] >> 3)
    b5 = (dec[v[8]] << 5) | dec[v[9]]
    # Entropy (16 characters -> 10 bytes) - fully unrolled
    b6 = (dec[v[10]] << 3) | (dec[v[11]] >> 2)
    b7 = (dec[v[11]] << 6) | (dec[v[12]] << 1) | (dec[v[13]] >> 4)
    b8 = (dec[v[13]] << 4) | (dec[v[14]] >> 1)
    b9 = (dec[v[14]] << 7) | (dec[v[15]] << 2) | (dec[v[16]] >> 3)
    b10 = (dec[v[16]] << 5) | dec[v[17]]
    b11 = (dec[v[18]] << 3) | (dec[v[19]] >> 2)
    b12 = (dec[v[19]] << 6) | (dec[v[20]] << 1) | (dec[v[21]] >> 4)
    b13 = (dec[v[21]] << 4) | (dec[v[22]] >> 1)
    b14 = (dec[v[22]] << 7) | (dec[v[23]] << 2) | (dec[v[24]] >> 3)
    b15 = (dec[v[24]] << 5) | dec[v[25]]

    data = bytes(
        [
            b0 & 0xFF,
            b1 & 0xFF,
            b2 & 0xFF,
            b3 & 0xFF,
            b4 & 0xFF,
            b5 & 0xFF,
            b6 & 0xFF,
            b7 & 0xFF,
            b8 & 0xFF,
            b9 & 0xFF,
            b10 & 0xFF,
            b11 & 0xFF,
            b12 & 0xFF,
            b13 & 0xFF,
            b14 & 0xFF,
            b15 & 0xFF,
        ]
    )
    return ULID(data)


def new(ms: int, entropy: EntropyReader = None) -> ULID:
    if ms > MAX_TIME:
        raise ErrBigTime()
    buf = bytearray(16)
    buf[0] = (ms >> 40) & 0xFF
    buf[1] = (ms >> 32) & 0xFF
    buf[2] = (ms >> 24) & 0xFF
    buf[3] = (ms >> 16) & 0xFF
    buf[4] = (ms >> 8) & 0xFF
    buf[5] = ms & 0xFF

    if entropy is not None:
        if hasattr(entropy, "monotonic_read"):
            rand_bytes = entropy.monotonic_read(ms)
            buf[6:] = rand_bytes[:10]
        elif callable(entropy):
            rand_bytes = entropy(10)
            buf[6:] = rand_bytes[:10]
        else:
            raise TypeError(f"unsupported entropy type: {type(entropy)}")

    return ULID(bytes(buf))


def must_new(ms: int, entropy: EntropyReader = None) -> ULID:
    return new(ms, entropy)


_default_entropy: EntropyReader = None
_default_entropy_initialized = False


def _get_default_entropy() -> EntropyReader:
    global _default_entropy, _default_entropy_initialized
    if not _default_entropy_initialized:
        from ulid.monotonic import LockedMonotonicReader, monotonic

        _default_entropy = LockedMonotonicReader(monotonic(os.urandom, 0))
        _default_entropy_initialized = True
    return _default_entropy


def make() -> ULID:
    return must_new(now(), _get_default_entropy())


def parse(s: str) -> ULID:
    return _parse(s, False)


def parse_strict(s: str) -> ULID:
    return _parse(s, True)


def parse_prefixed(s: str) -> tuple[str, ULID]:
    idx = s.find("_")
    if idx == -1:
        raise ErrInvalidPrefix()
    prefix = s[:idx]
    ulid_part = s[idx + 1 :]
    if len(ulid_part) != ENCODED_SIZE:
        raise ErrDataSize()
    return prefix, _parse(ulid_part, False)


def must_parse(s: str) -> ULID:
    return parse(s)


def must_parse_strict(s: str) -> ULID:
    return parse_strict(s)


def now() -> int:
    return int(_time.time() * 1000)


def timestamp(dt: datetime) -> int:
    return int(dt.timestamp() * 1000)


def time(ms: int) -> datetime:
    return datetime.fromtimestamp(ms / 1000.0, tz=timezone.utc)


def max_time() -> int:
    return MAX_TIME


def default_entropy() -> EntropyReader:
    return _get_default_entropy()


ZERO = ULID()



from ulid.generator import (
    Generator,
    GeneratorOption,
    new_generator,
    with_entropy,
    with_node_id,
    with_prefix,
)
from ulid.monotonic import (
    LockedMonotonicReader,
    MonotonicEntropy,
    monotonic,
)
from ulid.ulid import (
    BINARY_SIZE,
    ENCODED_SIZE,
    MAX_TIME,
    ULID,
    ZERO,
    EntropyReader,
    ErrBigTime,
    ErrBufferSize,
    ErrDataSize,
    ErrInvalidCharacters,
    ErrInvalidPrefix,
    ErrMonotonicOverflow,
    ErrOverflow,
    ErrScanValue,
    ULIDError,
    default_entropy,
    make,
    max_time,
    must_new,
    must_parse,
    must_parse_strict,
    new,
    now,
    parse,
    parse_prefixed,
    parse_strict,
    time,
    timestamp,
)

__all__ = [
    "BINARY_SIZE",
    "ENCODED_SIZE",
    "MAX_TIME",
    "ZERO",
    "ULID",
    "ULIDError",
    "ErrDataSize",
    "ErrInvalidCharacters",
    "ErrBufferSize",
    "ErrBigTime",
    "ErrOverflow",
    "ErrMonotonicOverflow",
    "ErrScanValue",
    "ErrInvalidPrefix",
    "EntropyReader",
    "new",
    "must_new",
    "make",
    "parse",
    "parse_strict",
    "parse_prefixed",
    "must_parse",
    "must_parse_strict",
    "now",
    "timestamp",
    "time",
    "max_time",
    "default_entropy",
    "MonotonicEntropy",
    "LockedMonotonicReader",
    "monotonic",
    "Generator",
    "GeneratorOption",
    "new_generator",
    "with_node_id",
    "with_entropy",
    "with_prefix",
]



from __future__ import annotations

import json
import os
import random
import threading
import time
from datetime import datetime, timezone

import ulid


class TestNew:
    def test_with_entropy(self) -> None:
        uid = ulid.new(ulid.now(), os.urandom)
        assert not uid.is_zero()

    def test_with_none_entropy(self) -> None:
        ms = ulid.now()
        uid = ulid.new(ms, None)
        assert uid.time() == ms
        assert uid.entropy() == b"\x00" * 10

    def test_with_big_time(self) -> None:
        try:
            ulid.new(ulid.max_time() + 1, None)
            assert False, "expected ErrBigTime"
        except ulid.ErrBigTime:
            pass

    def test_with_max_time(self) -> None:
        uid = ulid.new(ulid.max_time(), None)
        assert uid.time() == ulid.max_time()


class TestMustNew:
    def test_success(self) -> None:
        uid = ulid.must_new(ulid.now(), os.urandom)
        assert not uid.is_zero()

    def test_panics_on_error(self) -> None:
        try:
            ulid.must_new(ulid.max_time() + 1, None)
            assert False, "expected exception"
        except ulid.ErrBigTime:
            pass


class TestMake:
    def test_basic(self) -> None:
        uid = ulid.make()
        assert not uid.is_zero()

        n = ulid.now()
        ts = uid.time()
        assert ts <= n
        assert ts >= n - 1000


class TestParse:
    def test_valid(self) -> None:
        orig = ulid.make()
        s = orig.string()
        parsed = ulid.parse(s)
        assert orig == parsed

    def test_case_insensitive(self) -> None:
        orig = ulid.make()
        s = orig.string()
        upper = ulid.parse(s.upper())
        assert orig == upper

    def test_wrong_length(self) -> None:
        try:
            ulid.parse("short")
            assert False, "expected ErrDataSize"
        except ulid.ErrDataSize:
            pass

    def test_overflow(self) -> None:
        try:
            ulid.parse("80000000000000000000000000")
            assert False, "expected ErrOverflow"
        except ulid.ErrOverflow:
            pass

    def test_max_valid(self) -> None:
        uid = ulid.parse("7ZZZZZZZZZZZZZZZZZZZZZZZZZ")
        assert uid.time() == ulid.max_time()


class TestParseStrict:
    def test_valid(self) -> None:
        orig = ulid.make()
        ulid.parse_strict(orig.string())

    def test_invalid_characters(self) -> None:
        try:
            ulid.parse_strict("0000000000000000000000000i")
            assert False, "expected ErrInvalidCharacters"
        except ulid.ErrInvalidCharacters:
            pass

    def test_uppercase_valid(self) -> None:
        orig = ulid.make()
        ulid.parse_strict(orig.string().upper())


class TestMustParse:
    def test_success(self) -> None:
        orig = ulid.make()
        parsed = ulid.must_parse(orig.string())
        assert orig == parsed

    def test_panics_on_error(self) -> None:
        try:
            ulid.must_parse("bad")
            assert False, "expected exception"
        except ulid.ErrDataSize:
            pass


class TestMustParseStrict:
    def test_panics_on_error(self) -> None:
        try:
            ulid.must_parse_strict("0000000000000000000000000i")
            assert False, "expected exception"
        except ulid.ErrInvalidCharacters:
            pass


class TestPrefixed:
    def test_basic_prefix(self) -> None:
        uid = ulid.make()
        s = uid.prefixed("user")
        assert s.startswith("user_")
        assert len(s) == len("user_") + ulid.ENCODED_SIZE

    def test_txn_prefix(self) -> None:
        uid = ulid.make()
        s = uid.prefixed("txn")
        assert s.startswith("txn_")

    def test_ulid_portion_is_lowercase(self) -> None:
        uid = ulid.make()
        s = uid.prefixed("evt")
        ulid_part = s[len("evt_") :]
        assert ulid_part == ulid_part.lower()

    def test_prefixed_round_trip(self) -> None:
        uid = ulid.make()
        s = uid.prefixed("user")
        prefix, parsed = ulid.parse_prefixed(s)
        assert prefix == "user"
        assert parsed == uid


class TestParsePrefixed:
    def test_valid(self) -> None:
        uid = ulid.make()
        s = uid.prefixed("user")
        prefix, parsed = ulid.parse_prefixed(s)
        assert prefix == "user"
        assert parsed == uid

    def test_no_underscore(self) -> None:
        try:
            ulid.parse_prefixed("nounderscore")
            assert False, "expected ErrInvalidPrefix"
        except ulid.ErrInvalidPrefix:
            pass

    def test_wrong_ulid_length(self) -> None:
        try:
            ulid.parse_prefixed("user_short")
            assert False, "expected ErrDataSize"
        except ulid.ErrDataSize:
            pass

    def test_single_char_prefix(self) -> None:
        uid = ulid.make()
        s = uid.prefixed("x")
        prefix, parsed = ulid.parse_prefixed(s)
        assert prefix == "x"
        assert parsed == uid


class TestLowercaseOutput:
    def test_string_lowercase(self) -> None:
        for _ in range(100):
            uid = ulid.make()
            s = uid.string()
            assert s == s.lower()

    def test_json_lowercase(self) -> None:
        uid = ulid.make()
        data = uid.marshal_json()
        inner = json.loads(data)
        assert inner == inner.lower()

    def test_marshal_text_lowercase(self) -> None:
        uid = ulid.make()
        data = uid.marshal_text()
        assert data.decode() == data.decode().lower()


class TestTimestamp:
    def test_timestamp(self) -> None:
        dt = datetime.now(tz=timezone.utc)
        ms = ulid.timestamp(dt)
        expected = int(dt.timestamp() * 1000)
        assert ms == expected

    def test_time(self) -> None:
        ms = 1609459200000  # 2021-01-01 00:00:00 UTC
        t = ulid.time(ms)
        assert int(t.timestamp()) == 1609459200

    def test_max_time(self) -> None:
        expected = (1 << 48) - 1
        assert ulid.max_time() == expected

    def test_timestamp_round_trip(self) -> None:
        dt = datetime.now(tz=timezone.utc)
        ms = ulid.timestamp(dt)
        recovered = ulid.time(ms)
        assert abs(dt.timestamp() - recovered.timestamp()) < 0.001


class TestULIDMethods:
    def test_string(self) -> None:
        uid = ulid.make()
        s = uid.string()
        assert len(s) == ulid.ENCODED_SIZE

    def test_bytes(self) -> None:
        uid = ulid.make()
        b = uid.bytes()
        assert len(b) == ulid.BINARY_SIZE

    def test_time(self) -> None:
        ms = 1234567890123
        uid = ulid.new(ms, None)
        assert uid.time() == ms

    def test_entropy(self) -> None:
        uid = ulid.new(ulid.now(), os.urandom)
        e = uid.entropy()
        assert len(e) == 10

    def test_is_zero(self) -> None:
        zero = ulid.ULID()
        assert zero.is_zero()

        uid = ulid.make()
        assert not uid.is_zero()

    def test_compare(self) -> None:
        a = ulid.new(1000, None)
        b = ulid.new(2000, None)
        assert a.compare(b) < 0
        assert b.compare(a) > 0
        assert a.compare(a) == 0


class TestSetTimeEntropy:
    def test_set_time(self) -> None:
        uid = ulid.ULID()
        uid = uid.set_time(12345)
        assert uid.time() == 12345

        try:
            uid.set_time(ulid.max_time() + 1)
            assert False, "expected ErrBigTime"
        except ulid.ErrBigTime:
            pass

    def test_set_entropy(self) -> None:
        uid = ulid.ULID()
        e = bytes(range(1, 11))
        uid = uid.set_entropy(e)
        assert uid.entropy() == e

        try:
            uid.set_entropy(b"\x00" * 5)
            assert False, "expected ErrDataSize"
        except ulid.ErrDataSize:
            pass


class TestMarshal:
    def test_marshal_binary_round_trip(self) -> None:
        uid = ulid.make()
        data = uid.marshal_binary()
        assert len(data) == ulid.BINARY_SIZE
        parsed = ulid.ULID.unmarshal_binary(data)
        assert uid == parsed

    def test_unmarshal_binary_wrong_size(self) -> None:
        try:
            ulid.ULID.unmarshal_binary(b"\x01\x02\x03")
            assert False, "expected ErrDataSize"
        except ulid.ErrDataSize:
            pass

    def test_marshal_text_round_trip(self) -> None:
        uid = ulid.make()
        data = uid.marshal_text()
        assert len(data) == ulid.ENCODED_SIZE
        assert data.decode() == uid.string()
        parsed = ulid.ULID.unmarshal_text(data)
        assert uid == parsed

    def test_json_round_trip(self) -> None:
        uid = ulid.make()
        data = uid.marshal_json()
        parsed = ulid.ULID.unmarshal_json(data)
        assert uid == parsed

    def test_json_unmarshal_errors(self) -> None:
        try:
            ulid.ULID.unmarshal_json(b"notquoted")
            assert False, "expected ErrDataSize"
        except ulid.ErrDataSize:
            pass
        try:
            ulid.ULID.unmarshal_json(b'""')
            assert False, "expected ErrDataSize"
        except ulid.ErrDataSize:
            pass


class TestEncodingRoundTrip:
    def test_stress(self) -> None:
        for i in range(1000):
            orig = ulid.make()
            s = orig.string()
            parsed = ulid.parse(s)
            assert orig == parsed, f"iteration {i}: text round-trip failed"

            data = orig.marshal_binary()
            bin_parsed = ulid.ULID.unmarshal_binary(data)
            assert orig == bin_parsed, f"iteration {i}: binary round-trip failed"


class TestSortOrder:
    def test_lexicographic(self) -> None:
        ids = [ulid.make() for _ in range(100)]
        strs = [uid.string() for uid in ids]
        sorted_ids = sorted(ids)
        sorted_strs = sorted(strs)
        for i, uid in enumerate(sorted_ids):
            assert uid.string() == sorted_strs[i]

    def test_monotonic(self) -> None:
        entropy = ulid.monotonic(os.urandom, 0)
        ms = ulid.now()
        prev = ulid.new(ms, entropy)
        for i in range(1000):
            nxt = ulid.new(ms, entropy)
            assert nxt.compare(prev) > 0, f"iteration {i}: monotonic order violated"
            prev = nxt

    def test_monotonic_new_millisecond(self) -> None:
        entropy = ulid.monotonic(os.urandom, 0)
        id1 = ulid.new(1000, entropy)
        id2 = ulid.new(2000, entropy)
        assert id1 != id2
        assert id2.compare(id1) > 0

    def test_monotonic_overflow(self) -> None:
        max_entropy_data = bytes([0xFF] * 10 + [0x00] * 7 + [0x01])
        idx = 0

        def fake_reader(n: int) -> bytes:
            nonlocal idx
            result = max_entropy_data[idx : idx + n]
            idx += n
            if len(result) < n:
                result = result + b"\x00" * (n - len(result))
            return result

        entropy = ulid.monotonic(fake_reader, 1)
        ms = ulid.now()
        ulid.new(ms, entropy)
        try:
            ulid.new(ms, entropy)
            assert False, "expected ErrMonotonicOverflow"
        except ulid.ErrMonotonicOverflow:
            pass


class TestConcurrency:
    def test_concurrent_make(self) -> None:
        ids: list[ulid.ULID] = [ulid.ULID() for _ in range(1000)]
        errors: list[Exception] = []

        def worker(idx: int) -> None:
            try:
                ids[idx] = ulid.make()
            except Exception as e:
                errors.append(e)

        threads = [threading.Thread(target=worker, args=(i,)) for i in range(1000)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert not errors
        seen: set[ulid.ULID] = set()
        for uid in ids:
            assert not uid.is_zero()
            assert uid not in seen, f"duplicate ULID: {uid.string()}"
            seen.add(uid)


class TestGenerator:
    def test_basic(self) -> None:
        gen = ulid.new_generator()
        uid = gen.make()
        assert not uid.is_zero()

    def test_with_node_id(self) -> None:
        gen = ulid.new_generator(ulid.with_node_id(42))
        uid = gen.make()
        nid, ok = gen.node_id()
        assert ok
        assert nid == 42
        b = uid.bytes()
        assert b[6] == 0
        assert b[7] == 42

    def test_with_prefix(self) -> None:
        gen = ulid.new_generator(ulid.with_prefix("txn"))
        s = gen.make_prefixed()
        assert s.startswith("txn_")

    def test_override_prefix(self) -> None:
        gen = ulid.new_generator(ulid.with_prefix("txn"))
        s = gen.make_prefixed("user")
        assert s.startswith("user_")

    def test_no_prefix_panics(self) -> None:
        gen = ulid.new_generator()
        try:
            gen.make_prefixed()
            assert False, "expected RuntimeError"
        except RuntimeError:
            pass

    def test_generator_new(self) -> None:
        gen = ulid.new_generator(ulid.with_node_id(100))
        ms = ulid.now()
        uid = gen.new(ms)
        assert uid.time() == ms
        b = uid.bytes()
        assert b[6] == 0
        assert b[7] == 100

    def test_distributed_uniqueness(self) -> None:
        num_nodes = 10
        ids_per_node = 1000
        all_ids: list[ulid.ULID] = [ulid.ULID() for _ in range(num_nodes * ids_per_node)]
        errors: list[Exception] = []

        def node_worker(node: int) -> None:
            try:
                gen = ulid.new_generator(ulid.with_node_id(node))
                offset = node * ids_per_node
                for i in range(ids_per_node):
                    all_ids[offset + i] = gen.make()
            except Exception as e:
                errors.append(e)

        threads = [threading.Thread(target=node_worker, args=(n,)) for n in range(num_nodes)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert not errors
        seen: set[ulid.ULID] = set()
        for uid in all_ids:
            assert not uid.is_zero()
            assert uid not in seen, f"duplicate ULID across nodes: {uid.string()}"
            seen.add(uid)

    def test_generator_concurrent(self) -> None:
        gen = ulid.new_generator(ulid.with_node_id(1))
        ids: list[ulid.ULID] = [ulid.ULID() for _ in range(1000)]
        errors: list[Exception] = []

        def worker(idx: int) -> None:
            try:
                ids[idx] = gen.make()
            except Exception as e:
                errors.append(e)

        threads = [threading.Thread(target=worker, args=(i,)) for i in range(1000)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert not errors
        seen: set[ulid.ULID] = set()
        for uid in ids:
            assert uid not in seen, f"duplicate ULID: {uid.string()}"
            seen.add(uid)


class TestEdgeCases:
    def test_zero_ulid(self) -> None:
        s = ulid.ZERO.string()
        expected = "00000000000000000000000000"
        assert s == expected
        parsed = ulid.parse(expected)
        assert parsed == ulid.ZERO

    def test_known_values(self) -> None:
        uid = ulid.new(0, None)
        assert uid.string() == "00000000000000000000000000"
        max_id = ulid.parse("7zzzzzzzzzzzzzzzzzzzzzzzzz")
        assert max_id.time() == ulid.max_time()

    def test_default_entropy(self) -> None:
        e = ulid.default_entropy()
        assert e is not None
        uid = ulid.new(ulid.now(), e)
        assert not uid.is_zero()

    def test_timestamp_preservation(self) -> None:
        timestamps = [0, 1, 1000, int(time.time() * 1000), ulid.max_time() - 1, ulid.max_time()]
        for ms in timestamps:
            uid = ulid.new(ms, os.urandom)
            assert uid.time() == ms
            parsed = ulid.parse(uid.string())
            assert parsed.time() == ms

    def test_overflow_boundary(self) -> None:
        for c in "01234567":
            s = c + "0" * 25
            ulid.parse(s)

        overflow_cases = [
            "80000000000000000000000000",
            "90000000000000000000000000",
            "a0000000000000000000000000",
            "g0000000000000000000000000",
            "z0000000000000000000000000",
        ]
        for s in overflow_cases:
            try:
                ulid.parse(s)
                assert False, f"string {s!r} should overflow"
            except ulid.ErrOverflow:
                pass

    def test_monotonic_large_increment(self) -> None:
        entropy = ulid.monotonic(os.urandom, (1 << 64) - 1)
        ms = ulid.now()
        ulid.new(ms, entropy)
        try:
            ulid.new(ms, entropy)
        except ulid.ErrMonotonicOverflow:
            pass


class TestPythonSpecific:
    def test_rich_comparisons(self) -> None:
        a = ulid.new(1000, None)
        b = ulid.new(2000, None)
        assert a < b
        assert a <= b
        assert b > a
        assert b >= a
        assert a == a
        assert a != b

    def test_hash_and_set(self) -> None:
        uid = ulid.make()
        s: set[ulid.ULID] = {uid}
        assert uid in s
        parsed = ulid.parse(uid.string())
        assert parsed in s

    def test_int_conversion(self) -> None:
        uid = ulid.new(0, None)
        assert int(uid) == 0
        uid2 = ulid.make()
        assert int(uid2) > 0

    def test_bytes_conversion(self) -> None:
        uid = ulid.make()
        b = bytes(uid)
        assert len(b) == 16
        assert b == uid.bytes()

    def test_sorted(self) -> None:
        ids = [ulid.make() for _ in range(100)]
        sorted_ids = sorted(ids)
        for i in range(len(sorted_ids) - 1):
            assert sorted_ids[i] <= sorted_ids[i + 1]

    def test_locked_monotonic_reader(self) -> None:
        inner = ulid.monotonic(os.urandom, 0)
        locked = ulid.LockedMonotonicReader(inner)
        errors: list[Exception] = []

        ms = ulid.now()

        def worker() -> None:
            try:
                ulid.new(ms, locked)
            except Exception as e:
                errors.append(e)

        threads = [threading.Thread(target=worker) for _ in range(100)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert not errors

    def test_monotonic_with_random(self) -> None:
        rng = random.Random(42)

        def rng_reader(n: int) -> bytes:
            return bytes(rng.getrandbits(8) for _ in range(n))

        entropy = ulid.monotonic(rng_reader, 0)
        ms = ulid.now()
        prev = ulid.new(ms, entropy)
        for i in range(100):
            nxt = ulid.new(ms, entropy)
            assert nxt.compare(prev) > 0, f"monotonic order violated at iteration {i}"
            prev = nxt



from __future__ import annotations

import random
import threading
from collections.abc import Callable

from ulid.ulid import ErrMonotonicOverflow

_MASK_U16 = 0xFFFF
_MASK_U64 = 0xFFFFFFFFFFFFFFFF
_MAX_U32 = 0xFFFFFFFF


class _UInt80:
    __slots__ = ("hi", "lo")

    def __init__(self) -> None:
        self.hi: int = 0
        self.lo: int = 0

    def add(self, n: int) -> bool:
        lo = self.lo
        self.lo = (self.lo + n) & _MASK_U64
        if self.lo < lo:
            hi = self.hi
            self.hi = (self.hi + 1) & _MASK_U16
            return self.hi < hi
        return False

    def is_zero(self) -> bool:
        return self.hi == 0 and self.lo == 0

    def read_from(self, p: bytes | bytearray) -> None:
        self.hi = (p[0] << 8) | p[1]
        self.lo = (
            (p[2] << 56) | (p[3] << 48) | (p[4] << 40) | (p[5] << 32) | (p[6] << 24) | (p[7] << 16) | (p[8] << 8) | p[9]
        )

    def write_to(self, p: bytearray) -> None:
        p[0] = (self.hi >> 8) & 0xFF
        p[1] = self.hi & 0xFF
        p[2] = (self.lo >> 56) & 0xFF
        p[3] = (self.lo >> 48) & 0xFF
        p[4] = (self.lo >> 40) & 0xFF
        p[5] = (self.lo >> 32) & 0xFF
        p[6] = (self.lo >> 24) & 0xFF
        p[7] = (self.lo >> 16) & 0xFF
        p[8] = (self.lo >> 8) & 0xFF
        p[9] = self.lo & 0xFF


class MonotonicEntropy:
    __slots__ = ("_reader", "_ms", "_inc", "_entropy", "_rng")

    def __init__(self, reader: Callable[[int], bytes], inc: int) -> None:
        self._reader = reader
        self._inc = inc if inc != 0 else _MAX_U32
        self._ms: int = 0
        self._entropy = _UInt80()
        self._rng: random.Random | None = None
        if isinstance(reader, random.Random):
            self._rng = reader

    def monotonic_read(self, ms: int) -> bytes:
        if not self._entropy.is_zero() and self._ms == ms:
            inc = self._increment()
            if self._entropy.add(inc):
                raise ErrMonotonicOverflow()
            buf = bytearray(10)
            self._entropy.write_to(buf)
            return bytes(buf)

        rand_bytes = self._reader(10)
        self._ms = ms
        self._entropy.read_from(rand_bytes)
        return rand_bytes[:10]

    def read(self, n: int) -> bytes:
        return self._reader(n)

    def _increment(self) -> int:
        if self._rng is not None:
            return 1 + self._rng.randrange(self._inc)
        rand_bytes = self._reader(8)
        val = (
            (rand_bytes[0] << 56)
            | (rand_bytes[1] << 48)
            | (rand_bytes[2] << 40)
            | (rand_bytes[3] << 32)
            | (rand_bytes[4] << 24)
            | (rand_bytes[5] << 16)
            | (rand_bytes[6] << 8)
            | rand_bytes[7]
        )
        return 1 + (val % self._inc)


class LockedMonotonicReader:
    __slots__ = ("_inner", "_lock")

    def __init__(self, inner: MonotonicEntropy) -> None:
        self._inner = inner
        self._lock = threading.Lock()

    def monotonic_read(self, ms: int) -> bytes:
        with self._lock:
            return self._inner.monotonic_read(ms)

    def read(self, n: int) -> bytes:
        with self._lock:
            return self._inner.read(n)

    def __call__(self, n: int) -> bytes:
        return self.read(n)


def monotonic(entropy: Callable[[int], bytes], inc: int) -> MonotonicEntropy:
    return MonotonicEntropy(entropy, inc)
